/**
 * @function streamMix
 * @param {function}
 * @returns {function}
 */
'use strict'

const asleep = require('asleep')
const {EventEmitter} = require('events')
const {unlessProduction} = require('the-check')
const uuid = require('uuid')
const IOEvents = require('../constants/IOEvents')

class RemoteStream extends EventEmitter {
  constructor (streamName, socket) {
    super()
    this.streamName = streamName
    this.socket = socket
    this.sid = uuid.v4()
    this.tmieout = 30 * 1000
    this.closed = false
  }

  fireRemoteEvent (event, data) {
    const {sid, socket} = this
    socket.emit(event, {...data, sid})
  }

  listenToRemoteEvent (event, handler) {
    const {socket} = this
    const handlerWrap = (received) => {
      if (received.sid === this.sid) {
        handler(received)
      }
    }
    socket.on(event, handlerWrap)
  }

  on (event, handler) {
    unlessProduction(() => {
      const isValidEvent = [
        'open',
        'error',
        'close',
        'receive',
      ].includes(event)
      if (!isValidEvent) {
        throw new Error(`[TheClient.RemoteStream] Unknown event: ${event}`)
      }
    })
    return super.on(event, handler)
  }

  send (data) {
    this.fireRemoteEvent(IOEvents.STREAM_CHUNK, {chunk: data})
  }

  async close () {
    this.fireRemoteEvent(IOEvents.STREAM_CLOSE, {})
    await new Promise((resolve, reject) => {
      const timeoutTimer = setTimeout(() => {
        reject(new Error(`[TheClient] Stream close timeout`))
      }, this.tmieout)
      this.listenToRemoteEvent(IOEvents.STREAM_DID_CLOSE, () => {
        clearTimeout(timeoutTimer)
        resolve()
      })
    })
  }

  async open (params) {
    const {streamName} = this
    this.fireRemoteEvent(IOEvents.STREAM_OPEN, {params, streamName})
    this.listenToRemoteEvent(IOEvents.STREAM_CHUNK, ({chunk}) => {
      this.emit('receive', chunk)
    })
    this.listenToRemoteEvent(IOEvents.STREAM_DID_CLOSE, () => {
      this.closed = true
      this.emit('close')
    })
    this.listenToRemoteEvent(IOEvents.STREAM_ERROR, ({error, sid}) => {
      this.emit('error')
    })
    await new Promise((resolve, reject) => {
      const timeoutTimer = setTimeout(() => {
        reject(new Error(`[TheClient] Stream open timeout`))
      }, this.tmieout)
      this.listenToRemoteEvent(IOEvents.STREAM_DID_OPEN, () => {
        clearTimeout(timeoutTimer)
        resolve()
      })
    })
  }

  async wait () {
    await asleep(0)
    while (!this.closed) {
      await asleep(1)
    }
  }
}

/** @lends streamMix */
function streamMix (Class) {
  class StreamMixed extends Class {

    async openStream (name, params) {
      const {socket} = this
      const stream = new RemoteStream(name, socket)
      await stream.open(params)
      return stream
    }
  }

  return StreamMixed
}

module.exports = streamMix
